import time
from selenium.webdriver.common.keys import Keys
from bs4 import BeautifulSoup
import pandas as pd
from tqdm import tqdm
from market_research.scraper._scraper_scheme import Scraper


class Rurlsearcher(Scraper):
    def __init__(self, enable_headless=True):
        super().__init__(enable_headless=enable_headless)
        self.wait_time = 2
        self.model_dictionary = {"sony": {
            "oled": [
                'XR-77A95L', 'XR-65A75L', 'XR-65A95K', 'XR-65A90J', 'XR-65A80CL',
                'XR-77A80CL', 'XR-77A80K', 'XR-65A80CK', 'XR-42A90K', 'XR-55A80CL',
                'XR-83A80CL', 'XR-55A80CK', 'XR-65A80K', 'XR-83A90J', 'XR-48A90K',
                'XR-77A80L', 'XR-55A95K', 'XR-77A80CK', 'XR-55A80L', 'XR-55A80K',
                'XR-55A95L', 'XR-65A95L', 'XR-55A75L', 'XR-65A80L', 'XR-83A80L',
                'XR-55A90J', 'bravia8'
            ],
            "mini_led": [
                'XR-65X93CL', 'XR-75X93CL', 'XR-75X93L', 'XR-85X93L', 'XR-85X95L',
                'XR-65X93L', 'bravia7', 'bravia9'
            ],
            "lcd": ['XR-55X90CK', 'XR-65X90CK', 'XR-75X90CK', 'XR-85X90CK',
                    'KD-55X80CK', 'KD-65X80CK', 'KD-75X80CK', 'KD-85X80CK',
                    'KD-43X85K', 'KD-50X85K', 'KD-55X85K', 'KD-65X85K', 'KD-75X85K',
                    'KD-85X85K', 'XR-55X90CL', 'XR-65X90CL', 'XR-75X90CL',
                    'XR-85X90CL', 'XR-65X95K', 'XR-75X95K', 'XR-85X95K', 'XR-55X90K',
                    'XR-65X90K', 'XR-75X90K', 'XR-85X90K', 'XR-55X90L', 'XR-65X90L',
                    'XR-75X90L', 'XR-85X90L', 'XR-98X90L', 'KD-43X80K', 'KD-50X80K',
                    'KD-55X80K', 'KD-65X80K', 'KD-75X80K', 'KD-85X80K', 'XR-75Z9K',
                    'XR-85Z9K', 'KD-32W830K', 'KD-55X77CL', 'KD-65X77CL', 'KD-75X77CL',
                    'KD-85X77CL', 'KD-43X77L', 'KD-50X77L', 'KD-55X77L', 'KD-65X77L',
                    'KD-75X77L', 'KD-85X77L','bravia3',]
        },
        "lge": {
            "oled": ['lg g4', 'lg c4', 'lg b4', 
                     'lg g3', 'lg c3', 'lg b3',
                 
            ],
            "mini_led": ['lg qned90t 2024', 'lg qned85 2024'
              
            ],
            "lcd": ['lg ut7570 2024',]
        },
        "sse": {
            "oled": ['s95d', 's90d',  
                     's95c', 's90c', 
                 
            ],
            "mini_led": ['qn95d', 'qn90d',
                         'qn95c', 'qn90c']
              
        }
        
        
        }
     
     
     
     
     
     
     
        self.model_url_preset = {"sony":
            {
                
                "oled": [
                    'https://www.rtings.com/tv/reviews/sony/a90k-oled',
                    'https://www.rtings.com/tv/reviews/sony/a80l-a80cl-oled',
                    'https://www.rtings.com/tv/reviews/sony/a95l-oled',
                    'https://www.rtings.com/tv/reviews/sony/bravia-8-oled'
                ],
                "mini_led": [
                    'https://www.rtings.com/tv/reviews/sony/bravia-9-qled',
                    'https://www.rtings.com/tv/reviews/sony/bravia-7-qled',
                    'https://www.rtings.com/tv/reviews/sony/x95l',
                    'https://www.rtings.com/tv/reviews/sony/x93l-x93cl'
                ],
                "lcd": [
                    'https://www.rtings.com/tv/reviews/sony/x90l-x90cl',
                    'https://www.rtings.com/tv/reviews/sony/x77l-x77cl',
                    'https://www.rtings.com/tv/reviews/sony/x85k',
                    'https://www.rtings.com/tv/reviews/sony/x95k',
                    'https://www.rtings.com/tv/reviews/sony/x80k-x80ck',
                    'https://www.rtings.com/tv/reviews/sony/bravia-3'
                ]
            },
            "lge":
            {
                "oled": [
                        'https://www.rtings.com/tv/reviews/lg/g4-oled',
                        'https://www.rtings.com/tv/reviews/lg/c4-oled',
                        'https://www.rtings.com/tv/reviews/lg/b4-oled'
                ],
                "mini_led": [
                    'https://www.rtings.com/tv/reviews/lg/qned90t',
                    'https://www.rtings.com/tv/reviews/lg/qned85'
                ],
                "lcd": [
                    'https://www.rtings.com/tv/reviews/lg/ut7570-ut75-ut7590'
                ]
            }
        }
        






    def get_model_from_dictionary(self, maker: str = "sony", key_mode=False):
        if key_mode:
            return self.model_dictionary.keys()
        return self.model_dictionary.get(maker.lower())

    def get_url_from_model_preset(self, maker: str = "sony", key_mode=False):
        if key_mode:
            return self.model_url_preset.keys()
        return self.model_url_preset.get(maker.lower())

    def get_urls_from_web(self, keywords: list = None) -> list:
        urls_set = set()
        for keyword in tqdm(keywords):
            url = self._search_and_extract_url(search_query=keyword)
            if url is not None:
                urls_set.add(url)
        return list(urls_set)

    def get_urls_from_inputpath(self, intput_folder_path: str) -> list:
        self.set_data_path(intput_folder_path=intput_folder_path)

        urls = []
        file_list = self.intput_folder.glob('*')
        excel_files = [file for file in file_list if file.suffix in {'.xlsx', '.xls'}]
        for excel_file in excel_files:
            df = pd.read_excel(excel_file)
            urls.extend(df["urls"])
        return urls

    def _search_and_extract_url(self, search_query: str, base_url="https://www.rtings.com"):
        driver = self.web_driver.get_chrome()
        try:
            driver.get(base_url)
            search_input = driver.find_element("class name", "searchbar-input")
            search_input.send_keys(search_query)
            search_input.send_keys(Keys.RETURN)
            time.sleep(self.wait_time)
            page_source = driver.page_source
            soup = BeautifulSoup(page_source, 'html.parser')
            search_results = soup.find_all('div', class_='searchbar_results-main')
            extracted_urls = [result.find('a')['href'] for result in search_results]

            # 추출된 URL을 /로 분할하고, 검색 결과와 비교하여 일치하는 경우 반환
            for url in extracted_urls:
                if "tv/reviews" in url:
                    split_url = url.split('/')
                    if len(split_url) == 5:
                        return base_url + url
        finally:
            # 브라우저 종료
            driver.quit()
        return None



import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import re
from typing import Optional, Union
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from market_research.scraper._visualization_scheme import BaseVisualizer
import plotly.graph_objects as go  
import plotly.express as px
import streamlit as st

class Rvisualizer(BaseVisualizer):

    def __init__(self, df:pd.DataFrame=None, output_folder_path="results", maker_filter=None):
        super().__init__(output_folder_path=output_folder_path)
        self.output_folder_path =output_folder_path
        
        if df is None:
            json_path = 'https://raw.githubusercontent.com/xikest/research_market_tv/main/json/measurement_data.json'
            self.df = pd.read_json(json_path, orient='records', lines=True)
        else:
            self.df = df.copy()   
        
        self.data_detail_dict: dict = {}
        self.data_detail_df = None
        self.title_units_dict = {
            "HDR Brightness": "cd/m²",
            "SDR Brightness": "cd/m²",
            "Black Frame Insertion (BFI)": "Hz",
            "Black Uniformity": "%",
            "Color Gamut": "%",
            "Color Volume": "cd/m²",
            "Color Volume(ITP)": "%",
            "Flicker-Free": "Hz",
            "Gray Uniformity": "%",
            "HDR Brightness In Game Mode": "cd/m²",
            "Reflections": "%",
            "Response Time": "ms",
            "Stutter": "ms",
            "Variable Refresh Rate": "Hz",
            "Viewing Angle": "°"
        }
        
        import streamlit as st
        self.df = self._initialize_data(self.df)
        
        if maker_filter is not None:
            if isinstance(maker_filter, str):
                maker_filter = [maker_filter]
            
            filtered_mask = self.df['maker'].isin(maker_filter)
            if filtered_mask.sum() > 0:
                self.df = self.df[filtered_mask]


    def _initialize_data(self, df):
        
        def retrim(ds: pd.Series, mark: str = ","):
            return ds.str.replace(mark, "")
        
        try:
            df.loc[df['label'] == '1,000 cd/m² DCI P3 Coverage ITP', 'header'] = 'Color Volume(ITP)'
            df.loc[df['label'] == '10,000 cd/m² Rec 2020 Coverage ITP', 'header'] = 'Color Volume(ITP)'
            df.loc[df['label'] == 'Contrast', 'label'] = 'Contrast_'
            label_dict = {'1,000 cd/m² DCI P3 Coverage ITP': "DCI",
                        '10,000 cd/m² Rec 2020 Coverage ITP': "BT2020",
                        'Blue Luminance': "Blue",
                        'Cyan Luminance': "Cyan",
                        'Green Luminance': "Green",
                        'Magenta Luminance': "Magenta",
                        'Red Luminance': "Red",
                        'White Luminance': "White",
                        'Yellow Luminance': "Yellow"}

            value_dict = {"N/A": "0", "Inf ": "1000000"}

            for target, label in label_dict.items():
                df.loc[:, "label"] = df.label.map(lambda x: x.replace(target, label))
            for target, value in value_dict.items():
                df.loc[:, "result_value"] = df.result_value.map(lambda x: x.replace(target, value))

            trim_marks = ["cd/m²", ",", "%", "°", "K", "ms", "Hz", "dB", ": 1"]
            for trim_mark in trim_marks:
                try:
                    df.loc[:, "result_value"] = retrim(df["result_value"], trim_mark)
                except:
                    pass

            df = df[(df['category'] != 'Sound Quality') & (df['category'] != 'Smart Features') & (df['category'] != 'Inputs')]
            df = df[~df.score.isin([""])]  # score가 있는 데이터만 사용
            df['result_value'] = df['result_value'].astype(float)
            df["score"] = df.score.astype(float)
        except:
            pass
        return df









    def plotsns_facet_bar(self, column: str, col_wrap=3, height=4, facet_yticks: Optional[Union[dict, list]] = None,
                          facet_ylims: Optional[Union[dict, list]] = None, show_annot=True,
                          swap_mode: bool = False, save_plot_name:str=None):
        def get_data(column):
            
            def set_data_detail(column: str = None):
                
                def brightness_trim(df):
                    df_peak = df[df['label'].str.contains("Window") & df['label'].str.contains("Peak")]
                    df_peak.loc[:, "label"] = df_peak["label"].map(lambda x: int(x.split("%")[0].split(" ")[-1]))

                    df_peak = df_peak.sort_values(["model", "label"], ascending=True)
                    df_peak["label"] = df_peak.label.map(lambda x: str(x) + "%")
                    df_peak = df_peak.rename(columns={"label": "APL"})
                    return df_peak
                
                self._target_column = column
                data_df = self.df[self.df["header"] == column]
                data_df = data_df.sort_values(["maker", "product", "label"], ascending=False)
                data_df = data_df.pivot(index=["maker", "product"], columns="label", values='result_value')
                data_df.columns = data_df.columns.map(lambda x: str(x))
                data_df = data_df.reset_index()
                data_df.loc[:, 'model'] = data_df['maker'] + '_' + data_df['product']
                data_df = data_df.drop(["maker", "product"], axis=1)

                data_df = data_df.melt(id_vars=['model'], var_name='label', value_name=column)

                if "brightness" in column.lower():
                    data_df = brightness_trim(data_df)

                return data_df
            
            data_df = self.data_detail_dict.get(column)

            if data_df is None:
                data_df = set_data_detail(column)
                self.data_detail_dict.update({column: data_df})
                data_df = self.data_detail_dict.get(column)
            return data_df


        df = get_data(column)
        mode_dict = self._plot_mode(column, swap_mode)
        col_y = mode_dict.get("col_y")
        col_x = mode_dict.get("col_x")
        col_facet = mode_dict.get("col_facet")

        color_palette = sns.color_palette("Set2", n_colors=len(set(df[col_x])))

        sns.set(style="whitegrid")
        g = sns.FacetGrid(df, col=col_facet, col_wrap=col_wrap, height=height, sharey=False, sharex=False)
        g.map_dataframe(sns.barplot, x=col_x, y=col_y, hue=col_x, palette=color_palette, )
        g.set_axis_labels(col_x, col_y)

        if facet_yticks:
            if isinstance(facet_yticks, dict):
                for idx, ytick in facet_yticks.items():
                    g.axes.flat[int(idx)].yaxis.set_ticks(ytick)
            elif isinstance(facet_yticks, list):
                for ax in g.axes.flat:
                    ax.yaxis.set_ticks(facet_yticks)

        if facet_ylims:
            if isinstance(facet_ylims, dict):
                for idx, ylim in facet_ylims.items():
                    g.axes.flat[idx].set(ylim=ylim)
            elif isinstance(facet_ylims, tuple):
                for ax in g.axes.flat:
                    ax.set(ylim=facet_ylims)

        # g.set_xticklabels(rotation=90, horizontalalignment='right')
        for ax in g.axes.flat:
            for label in ax.get_xticklabels():

                if len(label.get_text()) > 5:
                    label.set_rotation(90)
                    label.set_horizontalalignment('right')
                else:
                    label.set_rotation(0)


        suffix = self.title_units_dict.get(col_y)
        if suffix is not None:
            sup_title = f"{col_y} ({suffix}) by {col_x}"
        else:
            sup_title = f"{col_y} by {col_x}"

        g.fig.suptitle(sup_title, y=1.05)
        g.fig.subplots_adjust(top=1)

        if show_annot:
            for ax in g.axes.flat:
                for p in ax.patches:
                    height_val = p.get_height()
                    annot_text = f'{height_val:.0f}' if height_val.is_integer() else f'{height_val}'
                    if height_val >= 10000:
                        annot_text = f'{height_val / 1000:.1f}K'
                    ax.annotate(annot_text, (p.get_x() + p.get_width() / 2., height_val), ha='center', va='center',
                                xytext=(0, 1), textcoords='offset points', fontsize=8)
                ax.yaxis.set_ticks([])
        sns.despine()
        plt.tight_layout()

        if save_plot_name is None:
            file_name = re.sub(r'\([^)]*\)', '', sup_title)
            save_plot_name = f"plot_for_{file_name}.png"
        plt.savefig(self.output_folder / save_plot_name, bbox_inches='tight')
        plt.show()

    import matplotlib.pyplot as plt

    def plot_lines(self, column, swap_mode=True, ylims: list = None,
                   yticks: list = None, save_plot_name: str = None):

        df = self._get_data(column)
        mode_dict = self._plot_mode(column, swap_mode)
        col_y = mode_dict.get("col_y")
        col_x = mode_dict.get("col_x")
        col_color = mode_dict.get("col_facet")

        suffix = self.title_units_dict.get(col_y)

        if suffix is not None:
            sup_title = f"{col_y} ({suffix}) by {col_x}"
        else:
            sup_title = f"{col_y} by {col_x}"

        fig, ax = plt.subplots(figsize=(10, 5))

        if col_color is not None:
            colors = plt.cm.viridis(np.linspace(0, 1, len(df[col_color].unique())))
            for i, (group, data) in enumerate(df.groupby(col_color)):
                ax.plot(data[col_x], data[col_y], label=group, color=colors[i])
                for x, y, val in zip(data[col_x], data[col_y], data[col_y]):
                    ax.annotate(f'{val}', (x, y), textcoords="offset points", xytext=(0, 10), ha='center')
        else:
            ax.plot(df[col_x], df[col_y])
            for x, y, val in zip(df[col_x], df[col_y], df[col_y]):
                ax.annotate(f'{val}', (x, y), textcoords="offset points", xytext=(0, 10), ha='center')

        ax.set_title(sup_title)
        ax.set_xlabel(col_x)
        ax.set_ylabel(col_y)

        if yticks is not None:
            ax.set_yticks(yticks)
        if ylims is not None:
            ax.set_ylim(ylims)

        if save_plot_name is None:
            file_name = re.sub(r'\([^)]*\)', '', sup_title)
            save_plot_name = f"plot_for_{file_name}.png"
        plt.savefig(save_plot_name)
        plt.show()


    def _plot_mode(self, column, swap_mode):
        col_y = column
        col_x = "model"
        col_facet = "label"

        if "brightness" in col_y.lower():
            col_x = "model"
            col_facet = "APL"

        if swap_mode == True:
            temp = col_x
            col_x = col_facet
            col_facet = temp

        return {"col_y": col_y,
                "col_x": col_x,
                "col_facet": col_facet}



    def radar_scores(self, return_fig: bool = False):
        
        def rgba_with_opacity(color, opacity=0.3):
            hex_color = color.lstrip('#')
            r, g, b = int(hex_color[0:2], 16), int(hex_color[2:4], 16), int(hex_color[4:6], 16)
            return f'rgba({r}, {g}, {b}, {opacity})'

        
        def get_measurement_data(df):
            col_socres = ["maker", "year", "series", "category", "header", "score"]
            data_df = df[col_socres].drop_duplicates().replace("", np.nan).dropna()
            
            data_df['year'] = data_df['year'].astype(int).astype(str)
            data_df["score"] = data_df["score"].map(lambda x: float(x))
            data_df = data_df.pivot(index=["maker", "year", "series"], columns=["category", "header"], values='score')
            data_df = data_df.T.reset_index().sort_index(axis=1, level=0).drop("category", axis=1).set_index("header")
            data_df = data_df.sort_index(axis='columns', level=[0, 1, 2], ascending=[True, False, False])


            data_df = data_df.loc[
                ['Black Uniformity', 'Contrast', 'SDR Brightness', 'HDR Brightness',
                'Color Volume', 'Color Gamut', 'Viewing Angle', 'Reflections',
                'Pre Calibration', 'Gray Uniformity', 'Response Time',
                'HDR Native Gradient', 'PQ EOTF Tracking'], :
            ]
            return data_df
        
        def get_scores_data(df):
            col_socres = ["maker", "year", "series", "category", "score"]
            data_df = df[col_socres].drop_duplicates().replace("", np.nan).dropna()

            data_df['year'] = data_df['year'].astype(int).astype(str)
            data_df["score"] = data_df["score"].map(lambda x: float(x))
            data_df = data_df.pivot(index=["maker", "year", "series"], columns=["category"], values='score')
            data_df = data_df.T.sort_index(axis='columns', level=[0, 1, 2], ascending=[True, False, False])
            return data_df
        
        
        ##data
        ##columns:"maker", "year", "series" ; multi columns
        ##index: header
        ## value: only keep one column

        # Plotly Figure 
    
        try:
            data_df = get_measurement_data(self.df)
        except:
            try:
                data_df = get_scores_data(self.df)
            except ValueError as e:
                print(e)
                
    
    
        fig = go.Figure()

        years = data_df.columns.get_level_values(1).unique()
        years = sorted(years, reverse=True)
        series_group = data_df.columns.get_level_values(2).unique()
        

        color_map = px.colors.qualitative.Plotly  # Plotly 색상 맵 사용
        
        
        
        for idx, series in enumerate(series_group):
            data_series = data_df.xs((series), level=(2), axis=1)
            year = data_series.columns.get_level_values(1).unique().item()
            maker = data_series.columns.get_level_values(0).unique().item()
            
            try:
                suffix = f": {data_series.loc['Mixed Usage'].item()}"
            except:
                suffix = ""
                        

            values = data_series.mean(axis=1).values.flatten()
            values = np.concatenate((values, [values[0]]))
            theta = data_df.index.tolist() + [data_df.index[0]]
            
            # 불연속 색상 맵에서 색상 선택
            line_color = color_map[idx % len(color_map)]  # 색상 맵의 색상 순환
            fill_color = rgba_with_opacity(line_color,opacity=0.1)
            fig.add_trace(go.Scatterpolar(
                r=values,
                theta=theta,
                fill='toself',
                fillcolor=fill_color,
                name=f"{maker} {series}({year}){suffix}",  # Series 이름에 연도 포함
                mode='lines',
                line=dict(color=line_color, width=2),
                visible= True  
            ))


        year_dropdown = dict(
            active=0,  
            buttons=[
                dict(
                    label='All',  
                    method='update',
                    args=[
                        {'visible': [True] * len(fig.data)}
                    ]
                )
            ] + [
                dict(
                    label=year,
                    method='update',
                    args=[
                        {'visible': [year in trace.name for trace in fig.data]}
                    ]
                ) for year in years
            ]
        )
    

        # 레이아웃 업데이트
        fig.update_layout(
            updatemenus=[{
                'buttons': year_dropdown['buttons'],
                'direction': 'down',
                'showactive': True,
                'x': 0,  # 가로 위치 (0~1)
                'y': 1.0,  # 세로 위치 (0~1) - 값을 조정하여 메뉴 위치 조정
                'yanchor': 'bottom'  # 기준점 설정
            }]
        )

        # 레이아웃 설정
        fig.update_layout(
            polar=dict(
                radialaxis=dict(
                    visible=False,
                    showline=False,
                    range=[0, 10],
                ),
                angularaxis=dict(
                    visible=True,
                    showline=True,
                    linecolor='gray',
                    linewidth=0.5,
                )
            ),
            showlegend=True,
            legend=dict(
                x=1.3,
                y=0.5,
                traceorder='normal',
                font=dict(size=12, family='Arial, sans-serif'),
            ),
            width=1000,
            height=800,
            font=dict(size=10, family='Arial, sans-serif', weight='bold'),
            margin=dict(l=80, r=100, t=10, b=10),
        )

        fig.write_html(self.output_folder / "radar_plot.html")
        if return_fig:
            return fig
        else:
            fig.show()  # 그래프 보여주기


 
    def heatmap_scores(self, cmap="cividis", return_fig:bool=False):
        col_socres = ["maker","year", "series", "category", "header", "score"]
        
        data_df = self.df[col_socres].drop_duplicates().replace("", np.nan).dropna()
        data_df["score"] = data_df["score"].map(lambda x: float(x))
        # data_df["product"] = data_df["product"].map(lambda x: x.replace("-oled", ""))
        data_df = data_df.pivot(index=["maker","year", "series"], columns=["category", "header"], values='score')
        # data_df = data_df.T.reset_index().sort_index(axis=1).drop("category", axis=1).set_index("header")
        data_df = data_df.T.reset_index().sort_index(axis=1, level=0).drop("category", axis=1).set_index("header")
        data_df = data_df.sort_index(axis=1, level=[0, 1])  # Sort the index levels
    

        x_labels = ['-'.join(map(str, idx)) for idx in data_df.columns]
        fig = go.Figure(data=go.Heatmap(
            z=data_df.values,  # 데이터 값
            x=x_labels,  # x축 레이블
            y=data_df.index,    # y축 레이블
            colorscale=cmap,     # 색상 맵
            showscale=False,
            zmin=0,              # 최소 값
            zmax=10,             # 최대 값
            text=data_df.values, # 각 셀의 텍스트로 데이터 값 표시
            texttemplate="%{text:.1f}",  # 텍스트 형식 지정
            hoverinfo='none'     # 호버 시 텍스트 정보 표시 안 함
        ))

        # 제목 설정
        fig.update_layout(
            title='Rtings Score heatmap',
            xaxis=dict(tickangle=90),
            yaxis=dict(tickangle=0),
            # width=1000,  # 그래프 너비
            height=1200,   # 그래프 높이
            font=dict(size=16, family='Arial, sans-serif', weight='bold'),
            margin=dict(l=300, r=20, t=40, b=300), 
        )

        fig.write_html(self.output_folder/f"heatmap.html")
                        
        if return_fig:
            return fig
        else:
            fig.show()  # 그래프 보여주기
        

    def plot_pca(self, figsize=(10, 6), title="Principal component", palette="RdYlBu", save_plot_name:str=None):
        sns.set_theme(style="whitegrid")
        ddf = self.df.copy()
        ddf['category_header_label'] = ddf['category'] + '_' + ddf['header'] + '_' + ddf['label']
        ddf['maker_product'] = ddf['maker'] + '_' + ddf['product']
        ddf = ddf.drop(["category", "header", "label", "score", "maker", "product"], axis=1)[
            ddf.category.isin(["Picture Quality"])]
    
        ddf_pivot = ddf.pivot_table(index=['maker_product'], columns=['category_header_label'],
                                    values=['result_value'],
                                    aggfunc={'result_value': 'first'})
        scaler = StandardScaler()
        X_numeric_scaled = scaler.fit_transform(ddf_pivot)
    
        pca = PCA(n_components=0.8)  # Set explained variance threshold to 0.8
        X_pca = pca.fit_transform(X_numeric_scaled)
    
        label_pc = [f"PC{i + 1}: {var*100:.2f}%" for i, var in enumerate(pca.explained_variance_ratio_)]

        pca_result_df = (
            pd.DataFrame(np.round(pca.components_.T * np.sqrt(pca.explained_variance_), 4),
                         columns=label_pc, index=ddf_pivot.columns)
            .reset_index()
            .assign(category=lambda x: x["category_header_label"].str.split("_").str[0],
                    header=lambda x: x["category_header_label"].str.split("_").str[1],
                    label=lambda x: x["category_header_label"].str.split("_").str[2])
            .drop("category_header_label", axis=1)
        )

        pca_result_by_header_df = pca_result_df.groupby(["header"])[label_pc].mean().reset_index()
        pca_result_by_header_df = pca_result_by_header_df.sort_values(by=label_pc[0], ascending=False)
        pca_result_long_df = pd.melt(pca_result_by_header_df, id_vars=["header"], var_name="Principal Component",
                                     value_name="loading")
    
        # 그래프 생성
        plt.figure(figsize=figsize)
        sns.barplot(x="loading", y="header", hue="Principal Component", data=pca_result_long_df, palette=palette)
        plt.xlim(-1, 1)
        plt.title(label=title, y=1.1)
        plt.legend(loc='upper center', bbox_to_anchor=(0.5, 1.1), fancybox=True, shadow=False, ncol=3)
    
        plt.tight_layout()
        sns.despine()
        if save_plot_name is None:
            save_plot_name = f"plot_pca_for_{title}.png"
        plt.savefig(self.output_folder / save_plot_name, bbox_inches='tight')
        plt.show()




from bs4 import BeautifulSoup
import time
from tqdm import tqdm
import pandas as pd
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.common.action_chains import ActionChains
from tools.file import FileManager
from market_research.scraper._scraper_scheme import Scraper, Modeler, CustomException
from market_research.scraper.models.visualizer.data_visualizer import DataVisualizer


class ModelScraper_s(Scraper, Modeler, DataVisualizer):
    def __init__(self, enable_headless=True,
                 export_prefix="sony_model_info_web", intput_folder_path="input", output_folder_path="results",
                 verbose: bool = False, wait_time=1, demo_mode:bool=False):

        Scraper.__init__(self, enable_headless=enable_headless, export_prefix=export_prefix, intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)

        self.tracking_log = verbose
        self.wait_time = wait_time
        self.file_manager = FileManager
        self.log_dir = "logs/sony/models"
        if self.tracking_log:
            FileManager.delete_dir(self.log_dir)
            FileManager.make_dir(self.log_dir)
            
        self._data = self._fetch_model_data(demo_mode=demo_mode)    
        DataVisualizer.__init__(self, df = self._data, plot_name='sony')
        pass
    
    def _fetch_model_data(self, demo_mode:bool=False) -> pd.DataFrame:
        def find_urls() -> dict:
            url_set = set()
            url_series_set = self._get_series_urls()
            for url in url_series_set:
                url_models_set = self._extract_models_from_series(url=url)
                url_set.update(url_models_set)
            url_dict = {idx: url for idx, url in enumerate(url_set)}
            print("total model:", len(url_dict))
            return url_dict
        
        def extract_specs(url_dict):
            dict_models = {}
            for key, url in tqdm(url_dict.items()):
                try:
                    dict_info = self._extract_model_details(url)
                    dict_models[key] = dict_info
                    dict_spec = self._extract_global_specs(url=url)
                    dict_models[key].update(dict_spec)
                except Exception as e:
                    pass
            return dict_models
        
        def transform_format(dict_models, json_file_name: str) -> pd.DataFrame:
            df_models = pd.DataFrame.from_dict(dict_models).T
            df_models.to_json(self.output_folder / json_file_name, orient='records', lines=True)
            return df_models
        
        if demo_mode:
            # Load existing JSON data
            df_models = pd.read_json('https://raw.githubusercontent.com/xikest/research_market_tv/main/json/s_scrape_model_data.json', orient='records', lines=True)
            print("operating demo")
        else:
            print("collecting models")
            url_dict = find_urls()
                
            dict_models = extract_specs(url_dict)
            
            df_models = transform_format(dict_models, json_file_name="s_scrape_model_data.json")
            
        FileManager.df_to_excel(df_models.reset_index(), file_name=self.output_xlsx_name)
        return df_models
    
    @Scraper.try_loop(2)
    def _get_series_urls(self) -> set:
        def find_series_urls(url:str, prefix:str) -> set:
            CustomException(message=f"error_find_series_urls: {url}")
            url_series = set()
            url = url
            prefix = prefix
            step = 200
            try:
                driver = self.web_driver.get_chrome()
                driver.get(url=url)
                time.sleep(self.wait_time)
                scroll_distance_total = self.web_driver.get_scroll_distance_total()
                scroll_distance = 0

                while scroll_distance < scroll_distance_total:
                    for _ in range(2):
                        html = driver.page_source
                        soup = BeautifulSoup(html, 'html.parser')
                        elements = soup.find_all('a', class_="custom-product-grid-item__image-container")
                        for element in elements:
                            url_series.add(prefix + element['href'].strip())
                        driver.execute_script(f"window.scrollBy(0, {step});")
                        time.sleep(self.wait_time)
                        scroll_distance += step
                        
                return url_series
            except CustomException as e:
                pass
            finally:
                driver.quit()
                
                            
        url_series = find_series_urls(url = "https://electronics.sony.com/tv-video/televisions/c/all-tvs/", prefix = "https://electronics.sony.com/")
        print("The website scan has been completed.")
        print(f"total series: {len(url_series)}")
        return url_series
    
    @Scraper.try_loop(2)
    def _extract_models_from_series(self, url: str) -> set:
        
        def extract_model_url(driver)->set:
            url_models_set= set()
            try: 
                elements = driver.find_element(By.CLASS_NAME,
                                               'custom-variant-selector__body')
            except:
                pass
                # try:
                #     elements = driver.find_element(By.XPATH,
                #                                         '//*[@id="PDPOveriewLink"]/div[1]/div/div/div[2]/div/app-custom-product-summary/div[2]/div/div[1]/app-custom-product-variants/div/app-custom-variant-selector/div/div[2]')
                # except:
                #     elements = driver.find_element(By.XPATH,
                #                                             '//*[@id="PDPOveriewLink"]/div[1]/div/div/div[2]/div/app-custom-product-summary/div/div/div[1]/app-custom-product-variants/div/app-custom-variant-selector/div/div[2]')
                    
            url_elements = elements.find_elements(By.TAG_NAME, 'a')
            
            for url_element in url_elements:
                url = url_element.get_attribute('href')
                url_models_set.add(url.strip())

            return url_models_set
        
        url_models_set = set()
        CustomException(message=f"error_extract_models_from_series: {url}")
        try:
            driver = self.web_driver.get_chrome()
            driver.get(url=url)
            time.sleep(self.wait_time)
            url_models_set = extract_model_url(driver)
        except CustomException as e:
            pass
        finally:
            driver.quit()
        return url_models_set
            
    @Scraper.try_loop(2)
    def _extract_model_details(self, url: str) -> dict:
        
        def extract_model(driver):
            model = driver.find_element(By.XPATH,
                                        '//*[@id="cx-main"]/app-product-details-page/div/app-custom-product-intro/div/div/div[1]/div/span').text

            model = model.split(":")[-1].strip()
            return {"model": model}
        
        def extract_description(driver)->dict:
            try:
                description = driver.find_element(By.XPATH,
                                                    '//*[@id="cx-main"]/app-product-details-page/div/app-custom-product-intro/div/div/div[1]/h1/p').text
            except:
                description = ""
            return {"description": description}
        
        def extract_prices(driver)->dict:
            prices_dict = dict()
            try:
                price_now = driver.find_element(By.XPATH,
                                                '//*[@id="PDPOveriewLink"]/div[1]/div/div/div[2]/div/app-custom-product-summary/app-product-pricing/div/div[1]/p[1]').text
                price_original  = driver.find_element(By.XPATH,
                                                        '//*[@id="PDPOveriewLink"]/div[1]/div/div/div[2]/div/app-custom-product-summary/app-product-pricing/div/div[1]/p[2]').text
                prices_dict['price'] = float(price_now.replace('$', '').replace(',', ''))
                prices_dict['price_original'] = float(price_original.replace('$', '').replace(',', ''))
                prices_dict['price_gap'] = price_original - price_now
            except:
                try:
                    price_now = driver.find_element(By.XPATH,
                                                    '//*[@id="PDPOveriewLink"]/div[1]/div/div/div[2]/div/app-custom-product-summary/app-product-pricing/div/div[1]/p').text
                    prices_dict['price'] = float(price_now.replace('$', '').replace(',', ''))
                    prices_dict['price_original'] = price_now
                    prices_dict['price_gap'] = 0.0
                except:
                    prices_dict['price'] = float('nan')
                    prices_dict['price_original'] = float('nan')
                    prices_dict['price_gap'] = float('nan')
            return prices_dict
         

        
        dict_info = {}
        CustomException(message=f"error_extract_model_details: {url}")
        try:
            driver = self.web_driver.get_chrome()
            driver.get(url)
            time.sleep(self.wait_time)
            
            # Extract model
            dict_info.update(extract_model(driver))
            dict_info.update(extract_description(driver))
            dict_info.update(extract_prices(driver))
            dict_info.update(ModelScraper_s.extract_info_from_model(dict_info.get("model")))
            
            if self.tracking_log:
                self._dir_model = f"{self.log_dir}/{dict_info['model']}"
                self.file_manager.make_dir(self._dir_model)
            return dict_info
        except CustomException as e:
            pass
        finally:
            driver.quit()
    
    @Scraper.try_loop(5)
    def _extract_global_specs(self, url: str) -> dict:
        def set_driver(url):
            driver = self.web_driver.get_chrome()
            driver.get(url=url)
            time.sleep(self.wait_time)
            return driver
   
        def find_spec_tab(driver) -> None:
            
            def find_element_spec(driver) -> None:
                element_spec = driver.find_element(By.XPATH, '//*[@id="PDPSpecificationsLink"]')
                self.web_driver.move_element_to_center(element_spec)
                if self.tracking_log:
                    driver.save_screenshot(f"./{self._dir_model}/{stamp_url}_1_move_to_spec_{stamp_today}.png")
                time.sleep(self.wait_time)
                return None
                
            def find_click_spec(driver) -> None:
                element_click_spec = driver.find_element(By.XPATH, '//*[@id="PDPSpecificationsLink"]/cx-icon')
                element_click_spec.click()
                time.sleep(self.wait_time)
                return None
            
            def remove_popup(driver) -> None:
                for _ in range(5):
                    try:
                        # 팝업 닫기 버튼을 기다렸다가 클릭
                        close_button = driver.find_element(By.XPATH, '//*[@id="contentfulModalClose"]')
                        close_button.click()
                        break
                    except Exception as e:
                        pass
                return None 

            def act_click_see_more(driver):
                try:
                    element_see_more = driver.find_element(By.XPATH,'//*[@id="cx-main"]/app-product-details-page/div/app-product-specification/div/div[2]/div[3]/button')
                    self.web_driver.move_element_to_center(element_see_more)
                    element_see_more.click()
                except:
                    try:
                        element_see_more = driver.find_element(By.XPATH, '//*[@id="cx-main"]/app-product-details-page/div/app-product-specification/div/div[2]/div[2]/button')
                        self.web_driver.move_element_to_center(element_see_more)
                        element_see_more.click()
                    except:
                        if self.tracking_log:
                            print("Cannot find the 'see more' button on the page")
                time.sleep(self.wait_time)
                return None
            
            find_element_spec(driver)
            find_click_spec(driver)
            if self.tracking_log: driver.save_screenshot( f"./{self._dir_model}/{stamp_url}_2_after_click_specification_{stamp_today}.png")
            remove_popup(driver)
            act_click_see_more(driver)    
            if self.tracking_log:driver.save_screenshot(f"./{self._dir_model}/{stamp_url}_3_after_click_see_more_{stamp_today}.png")
            
            return None
            
        def extract_specs(driver) -> dict:
            def convert_soup_to_dict(soup):
                """
                Convert BeautifulSoup soup to dictionary.
                """
                try:
                    h4_tag = soup.find('h4').text.strip()
                    p_tag = soup.find('p').text.strip()
                except :
                    try:
                        h4_tag = soup.find('h4').text.strip()
                        p_tag = ""
                    except Exception as e:
                        print("Parser error", e)
                        h4_tag = "Parser error"
                        p_tag = "Parser error"
                return h4_tag, p_tag
            
            dict_spec = {}
            driver.find_element(By.ID, "ngb-nav-0-panel").click()
            for _ in range(15):
                elements = driver.find_elements(By.CLASS_NAME,"full-specifications__specifications-single-card__sub-list")
                for element in elements:
                    soup = BeautifulSoup(element.get_attribute("innerHTML"), 'html.parser')
                    
                label, content  = convert_soup_to_dict(soup)
                original_label = label
                while label in dict_spec:
                    # 현재 item_name에서 *의 개수를 세어 그 개수에 따라 새로운 item_name 생성
                    asterisk_count = label.count('*')
                    label = f"{original_label}{'*' * (asterisk_count + 1)}"
                    
                dict_spec[label] = content
                
                
                ActionChains(driver).key_down(Keys.PAGE_DOWN).perform()

            return dict_spec
        
        dict_spec = {}
        CustomException(message=f"error_extract_global_specs: {url}")
        try:
            driver = set_driver(url)
            if self.tracking_log:
                stamp_today = self.file_manager.get_datetime_info(include_time=False)
                stamp_url = self.file_manager.get_name_from_url(url)
                driver.save_screenshot(f"./{self._dir_model}/{stamp_url}_0_model_{stamp_today}.png")
                
            find_spec_tab(driver)   
            dict_spec= extract_specs(driver)
            
            if self.tracking_log:
                driver.save_screenshot(f"./{self._dir_model}/{stamp_url}_4_end_{stamp_today}.png")
                print(f"Received information from {url}")
            return dict_spec
        except CustomException as e:
            pass
        finally:
            driver.quit()
            
    @staticmethod         
    def extract_info_from_model(model: str)->dict:
        dict_info = {}
        model = model.lower()
        dict_info["model"] = model
        dict_info["year"] = model.split("-")[1][-1]
        dict_info["series"] = model.split("-")[1][2:]
        dict_info["size"] = model.split("-")[1][:2]
        dict_info["grade"] = model.split("-")[0]
        
        
        year_mapping = {
            'l': "2023",
            'k': "2022",
            'j': "2021",
        }

        dict_info["year"] = year_mapping.get(dict_info.get("year"), "2024")

        return dict_info


import pandas as pd
import seaborn as sns
import os
import wget
from nltk.probability import FreqDist
from nltk import pos_tag
from wordcloud import WordCloud
import matplotlib.pyplot as plt
from nltk.tokenize import word_tokenize
from nltk.corpus import stopwords
import nltk


# import PyPDF2
import fitz  # PyMuPDF
from ._analysis_scheme import Analysis
class TextAnalysis(Analysis):
    def __init__(self,
                 export_prefix="text_", intput_folder_path="/content/input",  output_folder_path="results"):
        self.comments:[str,]
        self.nouns:[str,]
        self.df_word_freq:pd.DataFrame
        super().__init__(export_prefix=export_prefix,  intput_folder_path=intput_folder_path, output_folder_path=output_folder_path)

        # self._set_datapack()
    def read_files_from_inputpath(self, docs_type="excel"):

        file_type = {"excel":['.xlsx', '.xls'],
                     "pdf":[".pdf"]

                     }
        file_list = self.intput_folder.glob('*')

        files = [file for file in file_list if file.suffix in file_type.get(docs_type)]
        return files



    def download_pdfs(self, urls:list):
        # current_directory = os.getcwd()
        # target_folder = os.path.join(current_directory, intput_folder_path)
        # print(target_folder)
        # if not os.path.exists(target_folder):
        #     # print("sds")
        #     # current_directory = os.getcwd()
        #     # if not os.path.isabs(target_folder):
        #     #     target_folder = os.path.join(current_directory, target_folder)
        #     # if not os.path.exists(target_folder):
        #     os.makedirs(target_folder)
        #     print(f"Folder created at: {target_folder}")
        # else:
        #     print(f"Folder already exists at: {target_folder}")

        # if not os.path.exists(target_folder):
        #     print(f"Folder does not exist, will create: {target_folder}")
        #     os.makedirs(target_folder)
        #     print(f"Folder created: {target_folder}")
        # else:
        #     print(f"Folder already exists, no need to create: {target_folder}")

            #
            # target_folder = os.path.join(current_directory, target_folder)
            # os.makedirs(target_folder)
            # print(target_folder)
        for url in urls:
            try:
                filename = os.path.join(self.intput_folder, os.path.basename(url))
                if not os.path.exists(filename):
                    wget.download(url, filename)
            except Exception as e:
                print(f"Error downloading {url}: {str(e)}")

        print(f"downloaded in '{os.path.join(os.getcwd(), self.intput_folder)}'")

    def set_comments(self, comments: list, cleaning_words: list = None) -> None:
        self.comments = comments
        self.cleaning_words = cleaning_words
        self._prepare_nouns()
        self._prepare_word_freq()
        return None

    def save_df_freq_as_excel(self, file_name: str = "your_excel_file_name"):
        df = self.df_word_freq
        df.to_excel(self.output_folder / (file_name + "_freq.xlsx"), index=False)

    def _set_datapack(self) -> None:
        nltk.download('punkt')
        nltk.download('stopwords')
        nltk.download('averaged_perceptron_tagger')
        return None
    def _prepare_nouns(self) -> None:
        all_words = []
        for comment in self.comments:
            tokens = word_tokenize(comment)
            all_words.extend(tokens)
        stop_words = set(stopwords.words('english'))
        stop_words.update(self.cleaning_words)
        filtered_words = [word.lower() for word in all_words if
                          word.isalnum() and word.lower() not in stop_words]
        self._nouns = [word for (word, tag) in pos_tag(filtered_words) if tag.startswith('N')]
        return None
    def _prepare_word_freq(self) -> None:
        noun_counts = FreqDist(self._nouns)
        df_word_freq = pd.DataFrame(list(noun_counts.items()), columns=['Nouns', 'Frequency'])
        self.df_word_freq = df_word_freq.sort_values(by='Frequency', ascending=False)
        return None

    def plot_freq(self, num_dis: int = 10):
        try:
            df_word_freq=self.df_word_freq
            sns.set(style="white")
            top_words = df_word_freq.sort_values(by="Frequency", ascending=False).head(num_dis)
            plt.figure(figsize=(10, 4))
            sns.barplot(x='Nouns', y='Frequency', data=top_words)
            plt.title("Top 10 Words Frequency")
            plt.xticks(rotation=45)
            sns.despine()
            plt.show()

        except Exception as e:
            print(e)
            print("data not prepared")

    def plot_wordcloud(self):
        nouns = self._nouns
        try:
            sns.set(style="white")
            wordcloud = WordCloud(width=800, height=400, background_color="white").generate(" ".join(nouns))
            plt.figure(figsize=(10, 4))
            plt.imshow(wordcloud, interpolation='bilinear')
            plt.axis("off")
            plt.show()

        except Exception as e:
            print(e)
            print("data not prepared")

    def pdf_to_text(self, pdf_path):
        text = ""
        try:
            with fitz.open(pdf_path) as pdf_document:
                for page_number in range(pdf_document.page_count):
                    page = pdf_document[page_number]
                    text += page.get_text().lower().strip()

        except Exception as e:
            print(f"Error: {e}")

        return text

    # def pdf_to_text(self, pdf_path):
    #     text = ''
    #     with open(pdf_path, 'rb') as file:
    #         pdf_reader = PyPDF2.PdfReader(file)
    #         num_pages = len(pdf_reader.pages)
    #
    #         for page_num in range(num_pages):
    #             page = pdf_reader.pages[page_num]
    #             text += page.extract_text()
    #
    #     return text


